#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

class RWBuffer {
	pthread_mutex_t m;
	pthread_cond_t readers;
	pthread_cond_t writers;

	int stato;   // 0 : libero   1: lettura  2: scrittura
	int nlettori;
	int block_scritt;
	int block_lett;

public:
	RWBuffer();

	void request_read();
	void release_read();

	void request_write();
	void release_write();
}; 

RWBuffer mybuff;

RWBuffer::RWBuffer()
{
	pthread_mutex_init(&m, 0);
	pthread_cond_init(&readers, 0);
	pthread_cond_init(&writers, 0);

	stato = 0;
	nlettori = 0;
	block_scritt = 0;
	block_lett = 0;
}

void RWBuffer::request_read()
{
	pthread_mutex_lock(&m);
	
	if (stato == 2 || (stato == 1 && block_scritt > 0)) {
		block_lett++;
		pthread_cond_wait(&readers, &m);
		block_lett--;
	}

	stato = 1;
	nlettori++;

	pthread_mutex_unlock(&m);
}

void RWBuffer::release_read()
{
	pthread_mutex_lock(&m);

	nlettori--;

	if (nlettori == 0) {
		stato = 0;
		pthread_cond_signal(&writers);
	}

	pthread_mutex_unlock(&m);
}

void RWBuffer::request_write()
{
	pthread_mutex_lock(&m);

	while (stato != 0) {
		block_scritt++;
		pthread_cond_wait(&writers, &m);
		block_scritt--;
	}

	stato = 2;

	pthread_mutex_unlock(&m);
}

void RWBuffer::release_write()
{
	pthread_mutex_lock(&m);

	stato = 0;
	
	if (block_lett > 0) 
		pthread_cond_broadcast(&readers);
	else if (block_scritt > 0) 
		pthread_cond_signal(&writers);

	pthread_mutex_unlock(&m);
}

void *reader(void *)
{
	while(1) {
		printf("LETTORE: richiesta\n");
		mybuff.request_read();
		printf("LETTORE: sto leggendo\n");
		for (int i=0; i<1000000000; i++); 
		printf("LETTORE: rilascio\n");
		mybuff.release_read();
		sleep(rand()%2+1);
	}
}

void *writer(void *)
{
	while(1) {
		printf("SCRITTORE: richiesta\n");
		mybuff.request_write();
		printf("SCRITTORE: sto scrivendo\n");
		for (int i=0; i<1000000000; i++); 
		printf("SCRITTORE: rilascio\n");
		mybuff.release_write();
		sleep(rand()%2+1);
	}
}

#define N_LETT   5
#define N_SCRITT 2

int main()
{
	pthread_t lett[N_LETT];
	pthread_t scritt[N_SCRITT];

	for (int i=0; i<N_LETT; i++) 
		pthread_create(&lett[i], 0, reader, 0);

	for (int i=0; i<N_SCRITT; i++) 
		pthread_create(&scritt[i], 0, writer, 0);

	pause();
}






