#include <mqueue.h>
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>

class ServerData {
  mqd_t server_q;
  mqd_t client_q;
  int nthreads;

 public:
  ServerData(int n);
  ~ServerData();

  void body();
};


ServerData::ServerData(int n)
{
  struct mq_attr attr;

  nthreads = n;

  attr.mq_flags = 0;
  attr.mq_maxmsg = 5;
  attr.mq_msgsize = sizeof(int);

  server_q = mq_open("/server.mq",
  		  O_RDONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IXUSR,
  		  &attr);

  if (server_q < 0) {
    perror("Could not create server message queue");
    exit(-1);
  }

  client_q = mq_open("/client.mq",
		     O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IXUSR,
		     &attr);

  if (client_q < 0) {
    perror("Could not create client message queue");
    exit(-1);
  }

}

ServerData::~ServerData()
{
  mq_close(server_q);
  mq_close(client_q);
}


void ServerData::body()
{
  int data;
  unsigned prio;
  int ret;

  for (int i=0; i<nthreads; i++) {
    ret = mq_receive(server_q, (char *)&data, sizeof(int), &prio);
    if (ret < 0) {
      perror("Error in receive");
      exit(-1);
    }
    printf("message received from thread %d\n", i);    
  }

  for (int i=0; i<nthreads; i++) {
    ret = mq_send(client_q, (char *)&data, sizeof(int), prio);
    if (ret < 0) {
      perror("Error in send");
      exit(-1);
    }
    printf("message sent to the %d-th thread\n", i);
  }
}

int main()
{

  ServerData s(4);

  s.body();
}
