#include <pthread.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>

class SynchObj {
  pthread_mutex_t m;
  pthread_cond_t c;
  int nblocked;
  int nthreads;
public:
  SynchObj(int n);

  void synch();
};

SynchObj::SynchObj(int n)
{
  nthreads = n;
  nblocked = 0;
  pthread_mutex_init(&m, 0);
  pthread_cond_init(&c, 0);
}

void SynchObj::synch()
{
  pthread_mutex_lock(&m);

  nblocked++;

  if (nblocked == nthreads) {
    nblocked = 0;
    pthread_cond_broadcast(&c);
  }
  else 
    pthread_cond_wait(&c, &m);
  
  pthread_mutex_unlock(&m);
}

void *thread_body(void *arg)
{
  SynchObj *s = (SynchObj *)arg;

  printf("Before the synchronization point...\n");

  s->synch();

  printf("After the synchronization point!\n");
}


int main(int argc, char *argv[])
{
  pthread_t *array;
  int i;

  if (argc < 2) {
    printf("Usage: %s <nthreads>\n", argv[0]);
    exit(0);
  }
  int n = atoi(argv[1]);
  printf("Example with %d threads\n", n);
  
  array = new pthread_t[n];

  SynchObj my_synch(n);

  for(i=0; i<n; i++) 
    pthread_create(&array[i], 0, thread_body, (void *)(&my_synch));
  
  for (i=0; i<n; i++) 
    pthread_join(array[i], 0);

}
