#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>
#include <semaphore.h>
#include <unistd.h>

#define NCONS    5
#define NELEM    4

class CircularArray {
    int array[NELEM];

    sem_t mutex;
    sem_t producer;
    //int bProd;

    sem_t consumer[NCONS];
    //int bCons[NCONS];

    int counter[NELEM];
    int head;
    int tail[NCONS];

    pthread_t id[NCONS];

    int find(pthread_t tid);

public:
    CircularArray();
    void insert(int elem);
    void extract(int &elem);
};

CircularArray::CircularArray()
{
    int i;
    sem_init(&mutex, 0, 1);
    head = 0;
    for (i=0; i<NCONS; i++) {
	sem_init(&consumer[i], 0, 0);
        tail[i] = 0;
	id[i] = -1;
    }
    sem_init(&producer, 0, NELEM);

    for (i=0; i< NELEM; i++) 
 	counter[i] = NCONS;
}

int CircularArray::find(pthread_t tid)
{
    bool found = false; int index = -1;

    sem_wait(&mutex);

    for (int i=0; i<NCONS; i++) 
         if (id[i] == tid) {
            found = true;
            index = i;
            break;
         }
    if (!found) {
         for (int i=0; i<NCONS; i++) 
              if (id[i] == -1) {
                  id[i] = tid;
                  index = i;
                  break;
              }
    }

    sem_post(&mutex);
    return index;
}

void CircularArray::insert(int elem)
{
   sem_wait(&producer);
   sem_wait(&mutex);
   array[head] = elem;
   head = (head + 1) % NELEM;
   sem_post(&mutex);
   for (int i = 0; i< NCONS; i++) sem_post(&consumer[i]); 
}

void CircularArray::extract(int &elem)
{
    int index = find(pthread_self());

    sem_wait(&consumer[index]);
    sem_wait(&mutex);
    elem = array[ tail[index] ];
    counter[ tail[index] ]--;
    if (counter[ tail[index] ] == 0) {
	counter[ tail[index] ] = NELEM;
        sem_post(&producer);
    }

    tail[index] = (tail[index] + 1) % NELEM;
    sem_post(&mutex);
}

CircularArray ca;

/*------------------------------------*/

void *producer(void *arg)
{
   int i = 0;
   while (1) {
       printf("Producer: inserting ...\n");
       ca.insert(i++);
       printf("Producer: done!\n");
       //sleep(1);
   }
}

void *consumer(void *arg)
{
    int i = (int) arg;
    int elem;

    while(1) {
	printf("Consumer %d: extracting... \n", i);
	ca.extract(elem);
        printf("Consumer %d: extracted %d\n", i, elem);
	sleep(1);
    }
}

int main()
{
    pthread_t prod;
    pthread_t cons[NCONS];
    int i;

    pthread_create(&prod, 0, producer, 0);
    for(i=0; i<NCONS; i++) 
	pthread_create(&cons[i], 0, consumer, (void *)i);

    pause();
}
