#include <pthread.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>

struct Queue {
	int array[10];
	int nelem;
	int head;
	int tail;

	sem_t mutex;

	int state;           // 0 : condizione normale
			     // 1 : condizione di stop 
	
	sem_t block_prod;
	int n_block_prod;

	sem_t block_cons;
	int n_block_cons;

	sem_t block_mana;
	int n_block_mana;
} queue;

void init()
{
	queue.nelem = 0;
	queue.head = 0;
	queue.tail = 0;
	queue.state = 0;

	sem_init(&queue.mutex,0 ,1 );
	sem_init(&queue.block_cons, 0, 0);
	sem_init(&queue.block_prod, 0, 0);
	sem_init(&queue.block_mana, 0, 0);

	queue.n_block_prod = 0;
	queue.n_block_cons = 0;
	queue.n_block_mana = 0;
}

void insert(int elem)
{
	sem_wait(&queue.mutex);

	while (queue.nelem == 10 || queue.state) {
		queue.n_block_prod++;
		sem_post(&queue.mutex);
		sem_wait(&queue.block_prod);
		sem_wait(&queue.mutex);
	}

	/* A */ 
	queue.array[queue.head] = elem;
	queue.head = (queue.head + 1) % 10;
	queue.nelem++;
	if (queue.n_block_cons > 0) {
		queue.n_block_cons--;
		sem_post(&queue.block_cons);
	}
	sem_post(&queue.mutex);
}

void extract(int *elem)
{
	sem_wait(&queue.mutex);

	while (queue.nelem == 0) {
		queue.n_block_cons++;
		sem_post(&queue.mutex);
		sem_wait(&queue.block_cons);
		sem_wait(&queue.mutex);
	}

	*elem = queue.array[queue.tail];
	queue.tail = (queue.tail + 1) % 10;
	queue.nelem--;

	if (queue.state == 0 && queue.n_block_prod > 0) {
		queue.n_block_prod--;
		sem_post(&queue.block_prod);
	}
	else if (queue.state == 1 && queue.n_block_mana > 0) {
		queue.n_block_mana--;
		sem_post(&queue.block_mana);
	}

	sem_post(&queue.mutex);
}

void start()
{
	int i;

	sem_wait(&queue.mutex);

	if (queue.state == 1) {
		if (queue.nelem > 0) {
			queue.n_block_mana++;
			sem_post(&queue.mutex);
			sem_wait(&queue.block_mana);
			sem_wait(&queue.mutex);
		}
		queue.state = 0;

		while (queue.n_block_prod > 0) {
			queue.n_block_prod--;
			sem_post(&queue.block_prod);
		}
	}

	sem_post(&queue.mutex);
}

void stop()
{
	sem_wait(&queue.mutex);

	queue.state = 1;
	
	sem_post(&queue.mutex);
}

void *producer(void *arg)
{
  int counter = 0;
  while (1) {
    printf("PROD: inserting!\n");
    insert(counter);
    counter++;
    sleep(1);
  } 
}

void *consumer(void *arg)
{
  int data;
  while (1) {
    printf("CONS: extracting!\n");
    extract(&data);
    sleep(1);
  }
}

#define NCONS   2
#define NPROD   4

int main()
{
  int i;
  pthread_t prod_tid[NPROD];
  pthread_t cons_tid[NCONS];

  printf("Press 's' for stop, 'b' for start, 'q' to exit\n");

  init();

  for(i=0; i<NPROD; i++)
    pthread_create(&prod_tid[i], 0, producer, 0);

  for(i=0; i<NCONS; i++)
    pthread_create(&cons_tid[i], 0, consumer, 0);

  while(1) {
    char c;

    c = getchar();

    if (c == 's') {
      printf("Doing stop ...\n");
      stop();
      printf("Stop command done! \n");
    }
    else if (c == 'b') {
      printf("Start command! \n");
      start();
      printf("Start command completed! \n");
    }
    else if (c == 'q') exit(-1);
  }
}















