#include <pthread.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

class CircularArray {
  int *array;          // elements are here
  int size;
  int head, tail;      // index in the array
  int nelem;           // number of elements inside the array
  
  sem_t mutex;         // init 1
  sem_t prod, cons;    // blocks producers and consumers (init to 0)
  int bprod, bcons;    // blocked producers and consumers (init to 0)
  sem_t manager;       // init to 0
  bool bmanager;       // blocked manager (init to false)
  int state;           // state of the queue   1: start, 0: stop 

public:
  CircularArray(int size);

  void insert(int elem);
  void extract(int &elem);

  void start();
  void stop();
}; 

CircularArray::CircularArray(int s)
{
  size = s;
  array = new int[size];
  head = tail = 0;
  nelem = 0;
  bmanager = false;
  bprod = bcons = 0;
  state = 1;

  sem_init(&mutex, 0, 1);
  sem_init(&prod, 0, 0);
  sem_init(&cons, 0, 0);
  sem_init(&manager, 0, 0);
}


void CircularArray::insert(int elem)
{
  sem_wait(&mutex);
  
  if (state == 0 || nelem == size) {
    bprod++;
    sem_post(&mutex);
    sem_wait(&prod);
    bprod--;
  }

  array[head] = elem;
  head = (head + 1) % size;
  nelem++;

  if (bcons > 0) 
    sem_post(&cons);
  else 
    if (bprod > 0 && nelem < size) sem_post(&prod);
  else 
    sem_post(&mutex);
}

void CircularArray::extract(int &elem)
{
  sem_wait(&mutex);

  //if (state != 0) {
//	sem_post(&mutex);
//	return;
//  } 

  if (nelem == 0) {
    bcons++;
    sem_post(&mutex);
    sem_wait(&cons);
    bcons--;
  }

  elem = array[tail];
  tail = (tail + 1) % size;
  nelem--;

  if (state == 0 && bmanager && nelem == 0) {
    sem_post(&manager);
  }
  else if (state == 1 && bprod > 0) {
    sem_post(&prod);
  }
  else {
    sem_post(&mutex);
  }
}


void CircularArray::stop()
{
  sem_wait(&mutex);

  state = 0;

  sem_post(&mutex);
}

void CircularArray::start()
{
  int i;

  sem_wait(&mutex);

  if (nelem > 0) {
    bmanager = true;
    sem_post(&mutex);
    sem_wait(&manager);
    bmanager = false;
  }

  state = 1;

  if (bprod > 0) 
    sem_post(&prod);
  else sem_post(&mutex);
}


CircularArray ca(5);


void *producer(void *arg)
{
  int counter = 0;
  while (1) {
    printf("PROD: inserting!\n");
    ca.insert(counter);
    counter++;
    sleep(2);
  } 
}

void *consumer(void *arg)
{
  int data;
  while (1) {
    printf("CONS: extracting!\n");
    ca.extract(data);
    sleep(2);
  }
}


#define NCONS   2
#define NPROD   4

int main()
{
  int i;
  pthread_t prod_tid[NPROD];
  pthread_t cons_tid[NCONS];

  printf("Press 's' for stop, 'b' for start, 'q' to exit\n");

  for(i=0; i<NPROD; i++)
    pthread_create(&prod_tid[i], 0, producer, 0);

  for(i=0; i<NCONS; i++)
    pthread_create(&cons_tid[i], 0, consumer, 0);

  while(1) {
    char c;

    c = getchar();

    if (c == 's') {
      printf("Stop command! \n");
      ca.stop();
    }
    else if (c == 'b') {
      printf("Start command! \n");
      ca.start();
      printf("Start command completed! \n");
    }
    else if (c == 'q') exit(-1);
  }
}
