#include <pthread.h>
#include <semaphore.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

class CircularArray {
	sem_t m;
	
	sem_t prod;
	int block_prod;
	sem_t cons;
	int block_cons;
	sem_t manag;
	int block_manag;

	int nelem;
	int maxelem;
	int fermo;

	int head;
	int tail;

	int *array;

public:
  CircularArray(int size);

  void insert(int elem);
  void extract(int &elem);

  void start();
  void stop();
};

CircularArray::CircularArray(int size)
{
	array = new int[size];
	maxelem = size;
	head = tail = 0;

	fermo = 0; nelem = 0;
	sem_init(&m, 0, 1);

	sem_init(&prod, 0, 0);
	sem_init(&cons, 0, 0);
	sem_init(&manag, 0, 0);

	block_prod = 0;
	block_cons = 0;
	block_manag = 0;
}

void CircularArray::insert(int elem)
{
	sem_wait(&m);

	if (nelem == maxelem || fermo) {
		block_prod++;
		sem_post(&m);
		sem_wait(&prod);
		//sem_wait(&m);
	}

	array[head] = elem;
	head = (head + 1) % maxelem;

	nelem++;

	if (block_cons > 0) {
		block_cons--;
		sem_post(&cons);
	}
	else if (block_prod > 0 && nelem < maxelem) {
		block_prod--;
		sem_post(&prod);
	}
	else sem_post(&m);
}

void CircularArray::extract(int &elem)
{
	sem_wait(&m);

	if (nelem == 0) {
		block_cons++;
		sem_post(&m);
		sem_wait(&cons);
		//sem_wait(&m);
	}

	elem = array[tail];
	tail = (tail + 1) % maxelem;

	nelem --;

	if (fermo == 1 && nelem == 0 && block_manag > 0) {
		block_manag--;
		sem_post(&manag);
	}
	else if (fermo == 0 && block_prod > 0) {
		block_prod--;
		sem_post(&prod);
	} 
	else sem_post(&m);
}

void CircularArray::stop()
{
	sem_wait(&m);
	fermo = 1;
	sem_post(&m);	
}

void CircularArray::start()
{
	int i;
	sem_wait(&m);

	if (fermo == 1 && nelem > 0) {
		block_manag++;
		sem_post(&m);
		sem_wait(&manag);
		//sem_wait(&m);
	}

	fermo = 0;

	//i = 0;
	if (block_prod > 0 && i < maxelem) {
		block_prod --;
		sem_post(&prod);
	} 
	else sem_post(&m);	
}


CircularArray ca(5);


void *producer(void *arg)
{
  int counter = 0;
  while (1) {
    printf("PROD: inserting!\n");
    ca.insert(counter);
    counter++;
    sleep(2);
  } 
}

void *consumer(void *arg)
{
  int data;
  while (1) {
    printf("CONS: extracting!\n");
    ca.extract(data);
    sleep(2);
  }
}


#define NCONS   2
#define NPROD   4

int main()
{
  int i;
  pthread_t prod_tid[NPROD];
  pthread_t cons_tid[NCONS];

  printf("Press 's' for fermo, 'b' for start, 'q' to exit\n");

  for(i=0; i<NPROD; i++)
    pthread_create(&prod_tid[i], 0, producer, 0);

  for(i=0; i<NCONS; i++)
    pthread_create(&cons_tid[i], 0, consumer, 0);

  while(1) {
    char c;

    c = getchar();

    if (c == 's') {
      printf("fermo command! \n");
      ca.stop();
    }
    else if (c == 'b') {
      printf("Start command! \n");
      ca.start();
      printf("Start command completed! \n");
    }
    else if (c == 'q') exit(-1);
  }
}














