#include <iostream>
#include <pthread.h>
#include <semaphore.h>
#include "bridge.hpp"

#define MAXCARS    5

Bridge::Bridge()
{
  numCars = 0;
  sem_init(&mutex, 0, 1);
  sem_init(&blockShip, 0, 0);
  sem_init(&blockCars, 0, 0);
  requestOpen = false;
  open = false;
  nBlockedCars = 0;
  blockedShip = false;
}

void Bridge::request_open()
{
  sem_wait(&mutex);
  requestOpen = true;
  if (numCars > 0) {
    blockedShip = true;
    sem_post(&mutex);
    sem_wait(&blockShip);
  }
  open = true;
  sem_post(&mutex);
}

void Bridge::close()
{
  sem_wait(&mutex);
  open = false;
  requestOpen = false;
  if (nBlockedCars > 0) {
    nBlockedCars--;
    sem_post(&blockCars);
  }
  else sem_post(&mutex);
}

void Bridge::enter()
{
  sem_wait(&mutex);
  if (requestOpen || numCars == MAXCARS) {
    nBlockedCars ++;
    sem_post(&mutex);
    sem_wait(&blockCars);
  }
  numCars++;
  sem_post(&mutex);
}

void Bridge::exit()
{
  sem_wait(&mutex);
  numCars--;
  if (!requestOpen && nBlockedCars > 0) {
    nBlockedCars--;
    sem_post(blockCars);
  }
  else if (requestOpen && nCars == 0) {
    blockedShip = false;
    sem_post(&blockShip);
  }
  else sem_post(&mutex);
}