#include <mqueue.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include "requestqueue.h"

// the two barber seats
int seat[2];
// the two customers that are seated
struct req customer[2];
// the waiting room
RequestQueue room; 

// returns the seat number or -1 if they are all occupied
int allocate_seat(const struct req &r)
{
	int seatn = -1;
	if (seat[0] == 0) seatn = 0;
	else if (seat[1] == 0) seatn = 1;
	
	if (seatn != -1) {
		customer[seatn] = r;
		seat[seatn] = 1;
	}
	
	return seatn;
}

int free_seat(const struct req &r)
{
	int seatn = -1;
	if (seat[0] == 1 && strcmp(customer[0].client_name, r.client_name) == 0) {
		seatn = 0;
		seat[0] = 0;
	}
	else if (seat[1] == 1 && strcmp(customer[1].client_name, r.client_name) == 0) {
		seatn = 1;
		seat[1] = 0;
	}
	else {
		printf("No seat found, discarding exit request");
	}
	return seatn;
}

bool reply(char *client_name, int seat)
{
	mqd_t client_q;
	int ret;
	struct resp r;
	
	client_q = mq_open(client_name, O_WRONLY);
	if (client_q < 0) {
		perror("Could not open client queue, discarding request");
		return false;
	}	
	r.seat = seat;
	ret = mq_send(client_q, (char *)&r, sizeof(struct resp), 1);
	if (ret < 0) {
		perror("Could not send message to client, discarding request");
		return false;
	}
	return true;
}

int main()
{
	mqd_t server_q;
	int ret;
	struct req request;
	unsigned prio;
	struct mq_attr ma;

	ma.mq_flags = 0;
	ma.mq_maxmsg = 10;
	ma.mq_msgsize = sizeof(struct req);

	server_q = mq_open(BARBER_QUEUE_NAME,
  		  O_RDONLY | O_CREAT, S_IRUSR | S_IWUSR | S_IXUSR,
  		  &ma);

	if (server_q < 0) sys_err("could not open server queue");

	// init data structure
	seat[0] = seat[1] = 0; 

	while(1) {
		int seatn;
		ret = mq_receive(server_q, (char *)&request, sizeof(struct req), &prio);
		if (ret < 0) sys_err("receive error");
		printf("RECEIVED message from %s, type %d, prio %d\n", 
			request.client_name, request.type, prio);

		if (request.type == 0) {
			seatn = allocate_seat(request);
			if (seatn != -1) {
				printf("ALLOCATED SEAT %d to %s\n", seatn, request.client_name);
				reply(request.client_name, seatn);
			}
			else if (!room.insert(request)) {
				printf("ROOM IS FULL! SENT NO TO %s\n", request.client_name);
				reply(request.client_name, -1);
			}
			else {
				printf("CUSTOMER %s waiting in the ROOM\n", request.client_name);
			}
		}
		else if (request.type == 1) {
			printf("CUSTOMER %s exiting, freeing the seat\n", request.client_name);
			seatn = free_seat(request);
			if (seat >= 0) {
				if (room.extract(request)) {
					printf("CUSTOMER %s sits down on seat %d\n",
						request.client_name, seatn);
					seatn = allocate_seat(request);
					reply(request.client_name, seatn);
				}
			}			
		}
		else {
			printf("Type unknown, discarding request\n");
		}
	}

} 
