#include <semaphore.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

class Barbershop {
  int seat[2];      // 0: free   1: busy
  int nwaiting;     // number of customers in the waiting room;
  sem_t m;          // mutex semaphore
  sem_t b_sem[2];   // semaphore to block barbers;
  sem_t c_sem[2];   // semaphore to block customers being served
  sem_t w_sem;      // semaphore to block customers in the waiting room
 public:
  Barbershop();

  void wait_for_customers(int i);
  void free_seat(int i);
  int wait_for_seat();
  void wait_for_service(int i);
};

Barbershop::Barbershop() 
{
  int i;

  seat[0] = seat[1] = 0;
  nwaiting = 0;
  sem_init(&m, 0, 1);
  for (i=0; i<2; i++) {
    sem_init(&b_sem[i], 0, 0);
    sem_init(&c_sem[i], 0, 0);
  }
  sem_init(&w_sem, 0, 0);
}



void Barbershop::wait_for_customers(int i)
{
  sem_wait(&b_sem[i]);
}

void Barbershop::free_seat(int i)
{
  sem_wait(&m);
  seat[i] = 0;
  sem_post(&c_sem[i]);
  if (nwaiting > 0) sem_post(&w_sem);
  else sem_post(&m);
}

int Barbershop::wait_for_seat()
{
  int ret;

  sem_wait(&m);
  if (seat[0] && seat[1] && nwaiting == 5) {
    sem_post(&m);
    return -1;
  }

  if (seat[0] && seat[1]) {
    nwaiting++;
    sem_post(&m);
    sem_wait(&w_sem);
    nwaiting--;
  }

  if (seat[0] == 0) ret = 0;
  else ret = 1;

  seat[ret] = 1;
  sem_post(&b_sem[ret]);
  sem_post(&m);
  return ret;
}


void Barbershop::wait_for_service(int s)
{
  sem_wait(&c_sem[s]);
}


Barbershop bs;



void *barber(void *arg)
{
  int index = (int)arg;

  while(1) {
    bs.wait_for_customers(index);

    printf("Barber   %d: serving a customer\n", index);
    sleep(3);
    printf("Barber   %d: end of service\n", index);
    
    bs.free_seat(index);
  }
}

void *customer(void *arg)
{
  int k = (int) arg;

  printf("Customer %d: enters the barbershop\n", k);
  int seat = bs.wait_for_seat();
  if (seat == -1) {
    printf("Customer %d: barbershop full, exiting\n", k);
    return 0;
  }
  else {
    printf("Customer %d: being served,\n", k);
    bs.wait_for_service(seat);
    printf("Customer %d: finshed, exiting\n", k);
  } 
  return 0;
}


int main(int argc, char *argv[])
{
  pthread_t *array;
  pthread_t barbers[2];
  int i;

  if (argc < 2) {
    printf("Usage: %s <ncustomers>\n", argv[0]);
    exit(0);
  }

  int ncust = atoi(argv[1]);
  
  printf("creating %d customers\n", ncust);

  array = (pthread_t *)malloc(sizeof(pthread_t)*ncust);

  for(i=0; i<2; i++) {
    pthread_create(&barbers[i], 0, barber, (void *)i);
  }

  for (i=0; i<ncust; i++) {
    pthread_create(&array[i], 0, customer, (void *)i);
    sleep(1);
  }

  for (i=0; i<ncust; i++) {
    pthread_join(array[i], 0);
  }

  printf("FINISHED!\n");
}
