#include <iostream>
#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>

#define SLOW_CASHIER    0
#define FAST_CASHIER    1

using namespace std;

class Supermarket {
  sem_t mutex;
  sem_t queue[2];
  int nQueued[2];
  int busy[2];
public:
  Supermarket();
  int wait_for_paying(int nobjects);
  void exit_supermarket(int cashier);

  void print();

} sm;

Supermarket::Supermarket()
{
  for (int i=0; i<2; i++) {
    sem_init(&queue[i], 0, 0);
    nQueued[i] = 0;
    busy[i] = 0;
  }
  sem_init(&mutex, 0, 1);
}

int Supermarket::wait_for_paying(int nobjects)
{
  int type;
  int result;

  if (nobjects <= 5) type = FAST_CASHIER;
  else type = SLOW_CASHIER;

  sem_wait(&mutex);

  if ((type == SLOW_CASHIER && busy[SLOW_CASHIER] || 
       (type == FAST_CASHIER && busy[FAST_CASHIER] && busy[SLOW_CASHIER]))) {
    nQueued[type]++;
    sem_post(&mutex);
    sem_wait(&queue[type]);
    nQueued[type]--;
  }

  if (!busy[type]) {
    busy[type] = 1;
    result = type;
  }
  else if (type == FAST_CASHIER && !busy[SLOW_CASHIER]) {
    busy[SLOW_CASHIER] = 1;
    result = SLOW_CASHIER;
  }
  else {
    cerr << "THIS IS AN ERROR! I CANNOT BE HERE!" << endl;
    exit(-1);
  }

  sem_post(&mutex);
  return result;
}

void Supermarket::exit_supermarket(int cashier)
{
  sem_wait(&mutex);
  busy[cashier] = 0;

  if (nQueued[cashier] > 0) 
    sem_post(&queue[cashier]);
  else if (cashier == SLOW_CASHIER && nQueued[FAST_CASHIER] > 0)
    sem_post(&queue[FAST_CASHIER]);
  else 
    sem_post(&mutex);
}

void Supermarket::print()
{
  sem_wait(&mutex);
  cout << " ------------------- STATUS ---------------------" << endl;

  cout << "   FAST: " << (busy[FAST_CASHIER] ? "BUSY" : "FREE") << "   ENQUEUED " << nQueued[FAST_CASHIER] << endl; 
  cout << "   SLOW: " << (busy[SLOW_CASHIER] ? "BUSY" : "FREE") << "   ENQUEUED " << nQueued[SLOW_CASHIER] << endl; 

  cout << " ------------------------------------------------" << endl;
  sem_post(&mutex);
}


void *thread(void *arg)
{
  int cashier;
  int n = (int) arg;                 //number of objects

  cout << "CUSTOMER: Going to pay with " << n << " objects" << endl;

  sm.print();

  cashier = sm.wait_for_paying(n);

  cout << "CUSTOMER: " << n << " objects, paying at "  
       << ((cashier == FAST_CASHIER) ? "FAST" : "SLOW") << endl;

  sm.print();

  // takes some seconds to pay (depending on the # objects)
  sleep(2 + n/2);
  
  sm.exit_supermarket(cashier);

  cout << "CUSTOMER: " << n << " objects, left " 
       <<  ((cashier == FAST_CASHIER) ? "FAST" : "SLOW") << endl;;
}


#define NCUST    20
 
int main()
{
  pthread_t cust[NCUST];
  int i;
  int n;

  for (i=0; i<NCUST; i++) {
    n = 1 + rand() % 10;
    pthread_create(&cust[i], 0, thread, (void *) n);
    // one customer every 2 seconds
    sleep(2);
  }
  pause();
  
}
